/* Extracts a filesystem back from a compressed fs file */

#include <stdio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <asm/byteorder.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <zlib.h>
#include "compressed_loop.h"

struct compressed_block
{
	size_t size;
	void *data;
};

int main(int argc, char *argv[])
{
	int handle;
	struct cloop_head head;
	unsigned int i;
	unsigned char *buffer, *clear_buffer;
	loff_t offset[2]={0,0};

	if (argc != 2) {
		fprintf(stderr, "Need filename\n");
		exit(1);
	}

	handle = open(argv[1], O_RDONLY);
	if (handle < 0) {
		perror("Opening compressed file\n");
		exit(1);
	}

	if (read(handle, &head, sizeof(head)) != sizeof(head)) {
		perror("Reading compressed file header\n");
		exit(1);
	}
	head.block_size=ntohl(head.block_size);
	head.num_blocks=ntohl(head.num_blocks);
	if(head.cloop_version) {
		head.cloop_version=ntohl(head.cloop_version);
		head.property_flags=ntohl(head.property_flags);
		head.extension_bytes=ntohl(head.extension_bytes);
		i=lseek(handle, head.extension_bytes, SEEK_CUR);
	} else {
         if (!strncmp(head.preamble, "#!/bin/sh\n#V2.0 Format", 22))
         {
          fprintf(stderr, "old 2.0 format detected\n");
          head.property_flags=CLOOP_PROPERTY_OFFS64;
         }
        else if (!strncmp(head.preamble, "#!/bin/sh\n#V1.0 Format", 22))
         {
          fprintf(stderr, "old 1.0 format detected\n");
          head.property_flags=CLOOP_PROPERTY_OFFS64|CLOOP_PROPERTY_HOSTORDER;
         }
        else
         {
          fprintf(stderr, "old 0.6x format detected\n");
          head.property_flags=0;
         }
	}

	buffer = malloc(head.block_size + head.block_size/1000
			+ 12 + 4);
	clear_buffer = malloc(head.block_size);
	fprintf(stderr, "%u blocks of size %u. Preamble:\n%s\n", 
		head.num_blocks, head.block_size, head.preamble);

	for (i = 0; i < head.num_blocks+1; i++, offset[0]=offset[1]) {
		int currpos;
		unsigned long destlen = head.block_size;
		unsigned int size;
		uint32_t offset32;
		if(head.property_flags&CLOOP_PROPERTY_OFFS64) {
			read(handle, &offset[1], sizeof(loff_t));
			if(!(head.property_flags&CLOOP_PROPERTY_HOSTORDER)) offset[1]=__be64_to_cpu(offset[1]);
                } else {
			currpos=read(handle, &offset32, sizeof(offset32));
			offset[1]=ntohl(offset32);
		}
		if(head.property_flags&CLOOP_PROPERTY_DIFF) offset[1]+=offset[0];
		//fprintf(stderr, "%i %lli %lli\n", i, offset[0], offset[1]);
		if(!offset[0]) continue;
		currpos = lseek(handle, 0, SEEK_CUR);
		if (lseek(handle, offset[0], SEEK_SET) < 0) {
			fprintf(stderr, "lseek to %Lu: %s\n",
				offset[0], strerror(errno));
			exit(1);
		}
                
                size=offset[1]-offset[0];
		if (size > head.block_size + head.block_size/1000
		    + 12 + 4) {
			fprintf(stderr, 
				"Size %u for block %u (offset %Lu) too big\n",
				size, i, offset[0]);
			exit(1);
		}
		read(handle, buffer, size);
		if (lseek(handle, currpos, SEEK_SET) < 0) {
			perror("seeking");
			exit(1);
		}

		fprintf(stderr, "Block %u length %u => %lu\n",
			i, size, destlen);
		/*if (i == 3) {
			fprintf(stderr,
				"Block head:%02X%02X%02X%02X%02X%02X%02X%02X\n",
				buffer[0],
				buffer[1],
				buffer[2],
				buffer[3],
				buffer[4],
				buffer[5],
				buffer[6],
				buffer[7]);
			fprintf(stderr,
				"Block tail:%02X%02X%02X%02X%02X%02X%02X%02X\n",
				buffer[3063],
				buffer[3064],
				buffer[3065],
				buffer[3066],
				buffer[3067],
				buffer[3068],
				buffer[3069],
				buffer[3070]);
		}*/
		switch (uncompress(clear_buffer, &destlen,
				   buffer, size)) {
		case Z_OK:
			break;

		case Z_MEM_ERROR:
			fprintf(stderr, "Uncomp: oom block %u\n", i);
			exit(1);

		case Z_BUF_ERROR:
			fprintf(stderr, "Uncomp: not enough out room %u\n", i);
			exit(1);

		case Z_DATA_ERROR:
			fprintf(stderr, "Uncomp: input corrupt %u\n", i);
			exit(1);

		default:
			fprintf(stderr, "Uncomp: unknown error %u\n", i);
			exit(1);
		}
		if (destlen != head.block_size) {
			fprintf(stderr, "Uncomp: bad len %u (%lu not %u)\n", i,
				destlen, head.block_size);
			exit(1);
		}
		write(STDOUT_FILENO, clear_buffer, head.block_size);
	}
	return 0;
}
